@@ -35,7 +35,6 @@
 #include <limits.h>
 #include <unistd.h>
 
-
 extern char *prog_name;
 
 #ifndef TRUE
@@ -50,25 +49,34 @@ do { \
 	exit(EXIT_FAILURE); \
 } while (0)
 
-
-#define DEFAULT_PORT 6809
 #define DEFAULT_VOTES 1
 #define MAX_INTERFACES 10
+#define MAX_FORMAT_OPTS 10
 #define MAX_NODE_NAME_LEN 65
 #define MAX_MCAST_NAME_LEN 256
 #define MAX_PATH_LEN 256
 
+#define DEBUG_STARTUP_ONLY 32
+
+enum format_opt
+{
+	FMT_NONE,
+	FMT_ID,
+	FMT_NAME,
+	FMT_TYPE,
+	FMT_ADDR,
+};
 
 struct commandline
 {
 	int operation;
-	int num_multicasts;
-        int num_interfaces;
         int num_nodenames;
-	char *multicast_names[MAX_INTERFACES];
+	char *multicast_addr;
 	char *nodenames[MAX_INTERFACES];
         char *interfaces[MAX_INTERFACES];
 	char *override_nodename;
+	char *key_filename;
+	char *format_opts;
 	int votes;
 	int expected_votes;
 	int two_node;
@@ -79,7 +87,6 @@ struct commandline
 	int force;
         int verbose;
         int nodeid;
-	int cluster_id;
 	int timeout;
 	unsigned int config_version;
 
@@ -91,6 +98,9 @@ struct commandline
 	int clustername_opt;
 	int wait_opt;
 	int wait_quorate_opt;
+	int fence_opt;
+	int addresses_opt;
+	int noccs_opt;
 };
 typedef struct commandline commandline_t;
 
