@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -13,6 +13,7 @@
 
 #include <linux/sched.h>
 #include <linux/slab.h>
+#include <linux/vmalloc.h>
 #include <linux/smp_lock.h>
 #include <linux/spinlock.h>
 #include <asm/semaphore.h>
@@ -31,6 +32,8 @@
 #include "quota.h"
 #include "trans.h"
 
+static int gfs_commit_write(struct file *file, struct page *page,
+							unsigned from, unsigned to);
 /**
  * get_block - Fills in a buffer head with details about a block
  * @inode: The inode
@@ -104,7 +107,7 @@ get_block_noalloc(struct inode *inode, sector_t lblock,
 
 static int
 get_blocks(struct inode *inode, sector_t lblock,
-	   struct buffer_head *bh_result, int create)
+		   struct buffer_head *bh_result, int create)
 {
 	struct gfs_inode *ip = get_v2ip(inode);
 	int new = create;
@@ -125,7 +128,6 @@ get_blocks(struct inode *inode, sector_t lblock,
 	return 0;
 }
 
-#if 0
 /**
  * get_blocks_noalloc - 
  * @inode:
@@ -148,12 +150,11 @@ get_blocks_noalloc(struct inode *inode, sector_t lblock,
 		return error;
 
 	if (gfs_assert_withdraw(get_v2sdp(inode->i_sb),
-				!create || buffer_mapped(bh_result)))
+							!create || buffer_mapped(bh_result)))
 		return -EIO;
 
 	return 0;
 }
-#endif
 
 /**
  * gfs_writepage - Write complete page
@@ -271,13 +272,33 @@ gfs_readpage(struct file *file, struct page *page)
 {
 	struct gfs_inode *ip = get_v2ip(page->mapping->host);
 	struct gfs_sbd *sdp = ip->i_sbd;
+	struct gfs_holder *gh;
 	int error;
 
 	atomic_inc(&sdp->sd_ops_address);
 
-	if (gfs_assert_warn(sdp, gfs_glock_is_locked_by_me(ip->i_gl))) {
+	/* When gfs_readpage is called from the sys_madvise code through the 
+	 * readahead code, the inode glock is not held. In this case, we hold 
+	 * the inode glock, unlock the page and return AOP_TRUNCATED_PAGE. The
+	 * caller will then reload the page and call gfs_readpage again. We 
+	 * also add the flag GL_READPAGE to denote that the glock was held in
+	 * this function and if so, we unlock it before leaving this function
+	 */
+	gh = gfs_glock_is_locked_by_me(ip->i_gl);
+	if (!gh) {
+		gh = kmalloc(sizeof(struct gfs_holder), GFP_NOFS);
+		if (!gh)
+			return -ENOBUFS;
+		gfs_holder_init(ip->i_gl, LM_ST_SHARED, 
+				GL_READPAGE | LM_FLAG_ANY, gh);
 		unlock_page(page);
-		return -ENOSYS;
+		error = gfs_glock_nq(gh);
+		if (error) {
+			gfs_holder_uninit(gh);
+			kfree(gh);
+			goto out;
+		}
+		return AOP_TRUNCATED_PAGE;
 	}
 
 	if (!gfs_is_jdata(ip)) {
@@ -292,6 +313,11 @@ gfs_readpage(struct file *file, struct page *page)
 	if (unlikely(test_bit(SDF_SHUTDOWN, &sdp->sd_flags)))
 		error = -EIO;
 
+	if (gh->gh_flags & GL_READPAGE) { /* If we grabbed the glock here */
+		gfs_glock_dq_uninit(gh);
+		kfree(gh);
+	}
+out:
 	return error;
 }
 
@@ -321,6 +347,13 @@ gfs_prepare_write(struct file *file, struct page *page,
 
 	atomic_inc(&sdp->sd_ops_address);
 
+	/* We can't set commit_write in the structure in the declare         */
+	/* because if we do, loopback (loop.c) will interpret that to mean   */
+	/* it's okay to do buffered writes without locking through sendfile. */
+	/* This is a kludge to get around the problem with loop.c because    */
+	/* the upstream community rejected my changes to loop.c.             */
+	ip->gfs_file_aops.commit_write = gfs_commit_write;
+
 	if (gfs_assert_warn(sdp, gfs_glock_is_locked_by_me(ip->i_gl)))
 		return -ENOSYS;
 
@@ -389,6 +422,7 @@ gfs_commit_write(struct file *file, struct page *page,
 			goto fail;
 	}
 
+	ip->gfs_file_aops.commit_write = NULL;
 	return 0;
 
  fail:
@@ -438,7 +472,7 @@ gfs_bmap(struct address_space *mapping, sector_t lblock)
  * Returns: errno
  */
 
-static int
+static ssize_t
 gfs_direct_IO(int rw, struct kiocb *iocb, const struct iovec *iov,
 	      loff_t offset, unsigned long nr_segs)
 {
@@ -455,11 +489,17 @@ gfs_direct_IO(int rw, struct kiocb *iocb, const struct iovec *iov,
 		return -EINVAL;
 
 	if (rw == WRITE && !get_transaction)
-		gb = get_block_noalloc;
+		gb = get_blocks_noalloc;
 
-	return blockdev_direct_IO(rw, iocb, inode,
+	if (rw == WRITE)
+		return blockdev_direct_IO(rw, iocb, inode,
 				  inode->i_sb->s_bdev, iov,
 				  offset, nr_segs, gb, NULL);
+	else
+		return blockdev_direct_IO_no_locking(rw, iocb, inode,
+				  inode->i_sb->s_bdev, iov,
+				  offset, nr_segs, gb, NULL);
+
 }
 
 struct address_space_operations gfs_file_aops = {
@@ -467,7 +507,6 @@ struct address_space_operations gfs_file_aops = {
 	.readpage = gfs_readpage,
 	.sync_page = block_sync_page,
 	.prepare_write = gfs_prepare_write,
-	.commit_write = gfs_commit_write,
 	.bmap = gfs_bmap,
 	.direct_IO = gfs_direct_IO,
 };
