@@ -68,7 +68,6 @@ static spinlock_t gnbd_lock = SPIN_LOCK_UNLOCKED;
 
 static void gnbd_class_release(struct class_device *class_dev)
 {
-	printk("releasing gnbd class\n");
 	/* FIXME -- What the hell do I have to free up here */
 }
 
@@ -169,7 +168,7 @@ static ssize_t store_sectors(struct class_device *class_dev,
 	
 	if (down_trylock(&dev->do_it_lock))
 		return -EBUSY;
-	res = sscanf(buf, "%Lu\n", &size);
+	res = sscanf(buf, "%Lu\n", (unsigned long long int *)&size);
 	if (res != 1){
 		up(&dev->do_it_lock);
 		return -EINVAL;
@@ -267,21 +266,19 @@ static const char *gnbdcmd_to_ascii(int cmd)
 
 static void gnbd_end_request(struct request *req)
 {
-	int uptodate = (req->errors == 0) ? 1 : 0;
-	request_queue_t *q = req->q;
+	int error = req->errors ? -EIO : 0;
+	struct request_queue *q = req->q;
 	unsigned long flags;
 
 	dprintk(DBG_BLKDEV, "%s: request %p: %s\n", req->rq_disk->disk_name,
-			req, uptodate? "done": "failed");
+			req, error ? "failed" : "done");
 
-	if (!uptodate)
-		printk("%s %d called gnbd_end_request with and error\n",
+	if (error)
+		printk("%s %d called gnbd_end_request with an error\n",
 		       current->comm, current->pid);	
 	
 	spin_lock_irqsave(q->queue_lock, flags);
-	if (!end_that_request_first(req, uptodate, req->nr_sectors)) {
-		end_that_request_last(req, uptodate);
-	}
+	__blk_end_request(req, error, req->nr_sectors << 9);
 	spin_unlock_irqrestore(q->queue_lock, flags);
 }
 
@@ -289,7 +286,7 @@ static void gnbd_end_request(struct request *req)
  *  Send or receive packet.
  */
 static int sock_xmit(struct socket *sock, int send, void *buf, int size,
-		int msg_flags)
+		int msg_flags, int can_signal)
 {
 	mm_segment_t oldfs;
 	int result;
@@ -300,13 +297,12 @@ static int sock_xmit(struct socket *sock, int send, void *buf, int size,
 
 	oldfs = get_fs();
 	set_fs(get_ds());
-	/* Allow interception of SIGKILL only
-	 * Don't allow other signals to interrupt the transmission */
 	spin_lock_irqsave(&current->sighand->siglock, flags);
 	oldset = current->blocked;
 	sigfillset(&current->blocked);
-	sigdelsetmask(&current->blocked, sigmask(SIGKILL) | sigmask(SIGTERM) |
-	              sigmask(SIGHUP));
+	if (can_signal)
+		sigdelsetmask(&current->blocked, sigmask(SIGKILL) |
+			      sigmask(SIGTERM) | sigmask(SIGHUP));
 	recalc_sigpending();
 	spin_unlock_irqrestore(&current->sighand->siglock, flags);
 
@@ -328,13 +324,9 @@ static int sock_xmit(struct socket *sock, int send, void *buf, int size,
 		else
 			result = sock_recvmsg(sock, &msg, size, 0);
 
-		if (signal_pending(current)) {
-			siginfo_t info;
-			spin_lock_irqsave(&current->sighand->siglock, flags);
-			printk(KERN_WARNING "gnbd (pid %d: %s) got signal %d\n",
-				current->pid, current->comm, 
-				dequeue_signal(current, &current->blocked, &info));
-			spin_unlock_irqrestore(&current->sighand->siglock, flags);
+		if (can_signal && signal_pending(current)) {
+			printk(KERN_WARNING "gnbd (pid %d: %s) got signal\n",
+				current->pid, current->comm);
 			result = -EINTR;
 			break;
 		}
@@ -358,23 +350,24 @@ static int sock_xmit(struct socket *sock, int send, void *buf, int size,
 }
 
 static inline int sock_send_bvec(struct socket *sock, struct bio_vec *bvec,
-		int flags)
+		int flags, int can_signal)
 {
 	int result;
 	void *kaddr = kmap(bvec->bv_page);
 	result = sock_xmit(sock, 1, kaddr + bvec->bv_offset, bvec->bv_len,
-			flags);
+			flags, can_signal);
 	kunmap(bvec->bv_page);
 	return result;
 }
 
 
-#define gnbd_send_req(dev, req) __gnbd_send_req((dev), (dev)->sock, (req))
+#define gnbd_send_req(dev, req, can_sig) \
+__gnbd_send_req((dev), (dev)->sock, (req), (can_sig))
 	
 int __gnbd_send_req(struct gnbd_device *dev, struct socket *sock,
-		struct request *req)
+		struct request *req, int can_signal)
 {
-	int result, i, flags;
+	int result, flags;
 	struct gnbd_request request;
 	unsigned long size = req->nr_sectors << 9;
 
@@ -399,7 +392,8 @@ int __gnbd_send_req(struct gnbd_device *dev, struct socket *sock,
 			(unsigned long long)req->sector << 9,
 			req->nr_sectors << 9);
 	result = sock_xmit(sock, 1, &request, sizeof(request),
-			(gnbd_cmd(req) == GNBD_CMD_WRITE)? MSG_MORE: 0);
+			(gnbd_cmd(req) == GNBD_CMD_WRITE)? MSG_MORE: 0,
+			can_signal);
 	if (result < 0) {
 		printk(KERN_ERR "%s: Send control failed (result %d)\n",
 				dev->disk->disk_name, result);
@@ -407,27 +401,26 @@ int __gnbd_send_req(struct gnbd_device *dev, struct socket *sock,
 	}
 
 	if (gnbd_cmd(req) == GNBD_CMD_WRITE) {
-		struct bio *bio;
+		struct req_iterator iter;
+		struct bio_vec *bvec;
 		/*
 		 * we are really probing at internals to determine
 		 * whether to set MSG_MORE or not...
 		 */
-		rq_for_each_bio(bio, req) {
-			struct bio_vec *bvec;
-			bio_for_each_segment(bvec, bio, i) {
-				flags = 0;
-				if ((i < (bio->bi_vcnt - 1)) || bio->bi_next)
-					flags = MSG_MORE;
-				dprintk(DBG_TX, "%s: request %p: sending %d bytes data\n",
-						dev->disk->disk_name, req,
-						bvec->bv_len);
-				result = sock_send_bvec(sock, bvec, flags);
-				if (result < 0) {
-					printk(KERN_ERR "%s: Send data failed (result %d)\n",
-							dev->disk->disk_name,
-							result);
-					goto error_out;
-				}
+		rq_for_each_segment(bvec, req, iter) {
+			flags = 0;
+			if (!rq_iter_last(req, iter))
+				flags = MSG_MORE;
+			dprintk(DBG_TX, "%s: request %p: sending %d bytes data\n",
+					dev->disk->disk_name, req,
+					bvec->bv_len);
+			result = sock_send_bvec(sock, bvec, flags,
+						can_signal);
+			if (result < 0) {
+				printk(KERN_ERR "%s: Send data failed (result %d)\n",
+						dev->disk->disk_name,
+						result);
+				goto error_out;
 			}
 		}
 	}
@@ -459,7 +452,7 @@ static inline int sock_recv_bvec(struct socket *sock, struct bio_vec *bvec)
 	int result;
 	void *kaddr = kmap(bvec->bv_page);
 	result = sock_xmit(sock, 0, kaddr + bvec->bv_offset, bvec->bv_len,
-			MSG_WAITALL);
+			MSG_WAITALL, 1);
 	kunmap(bvec->bv_page);
 	return result;
 }
@@ -467,21 +460,19 @@ static inline int sock_recv_bvec(struct socket *sock, struct bio_vec *bvec)
 int gnbd_recv_req(struct gnbd_device *dev, struct request *req)
 {
 	int result;
-	int i;
-	struct bio *bio;
-	rq_for_each_bio(bio, req) {
-		struct bio_vec *bvec;
-		bio_for_each_segment(bvec, bio, i) {
-			result = sock_recv_bvec(dev->sock, bvec);
-			if (result < 0) {
-				printk(KERN_ERR "%s: Receive data failed (result %d)\n",
-						dev->disk->disk_name,
-						result);
-				return result;
-			}
-			dprintk(DBG_RX, "%s: request %p: got %d bytes data\n",
-					dev->disk->disk_name, req, bvec->bv_len);
+	struct bio_vec *bvec;
+	struct req_iterator iter;
+
+	rq_for_each_segment(bvec, req, iter) {
+		result = sock_recv_bvec(dev->sock, bvec);
+		if (result < 0) {
+			printk(KERN_ERR "%s: Receive data failed (result %d)\n",
+					dev->disk->disk_name,
+					result);
+			return result;
 		}
+		dprintk(DBG_RX, "%s: request %p: got %d bytes data\n",
+				dev->disk->disk_name, req, bvec->bv_len);
 	}
 	return 0;
 }
@@ -495,7 +486,7 @@ int gnbd_do_it(struct gnbd_device *dev)
 
 	BUG_ON(dev->magic != GNBD_MAGIC);
 
-	while((result = sock_xmit(sock, 0, &reply,sizeof(reply), MSG_WAITALL)) > 0){
+	while((result = sock_xmit(sock, 0, &reply,sizeof(reply), MSG_WAITALL, 1)) > 0){
 		if (ntohl(reply.magic) == GNBD_KEEP_ALIVE_MAGIC)
 			/* FIXME -- I should reset the wait time here */
 			continue;
@@ -568,7 +559,7 @@ void gnbd_clear_que(struct gnbd_device *dev)
  *   { printk( "Warning: Ignoring result!\n"); gnbd_end_request( req ); }
  */
 
-static void do_gnbd_request(request_queue_t * q)
+static void do_gnbd_request(struct request_queue *q)
 {
 	int err;
 	struct request *req;
@@ -582,7 +573,7 @@ static void do_gnbd_request(request_queue_t * q)
 
 		if (!blk_fs_request(req))
 			goto error_out;
-		
+
 		dev = req->rq_disk->private_data;
 
 		if (dev->receiver_pid == -1)
@@ -610,7 +601,7 @@ static void do_gnbd_request(request_queue_t * q)
 		list_add(&req->queuelist, &dev->queue_head);
 		spin_unlock(&dev->queue_lock);
 
-		err = gnbd_send_req(dev, req);
+		err = gnbd_send_req(dev, req, 0);
 
 		spin_lock_irq(q->queue_lock);
 		if (err)
@@ -642,7 +633,7 @@ static int gnbd_resend_requests(struct gnbd_device *dev, struct socket *sock)
 	printk("resending requests\n");
 	list_for_each(tmp, &dev->queue_head) {
 		req = list_entry(tmp, struct request, queuelist);
-		err = __gnbd_send_req(dev, sock, req);
+		err = __gnbd_send_req(dev, sock, req, 1);
 
 		if (err){
 			printk("failed trying to resend request (%d)\n", err);
@@ -706,7 +697,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		/* There is no one using the device, you can disconnect it */
 		if (dev->sock == NULL)
 			return -ENOTCONN;
-		gnbd_send_req(dev, &shutdown_req);
+		gnbd_send_req(dev, &shutdown_req, 1);
                 return 0;
 	case GNBD_CLEAR_QUE:
 		if (down_interruptible(&dev->do_it_lock))
@@ -716,8 +707,10 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		bdev = dev->bdev;
 		if (bdev) {
 			blk_run_queue(dev->disk->queue);
+#if 0
 			fsync_bdev(bdev);
-			invalidate_bdev(bdev, 0);
+			invalidate_bdev(bdev);
+#endif
 		}
 		up(&dev->do_it_lock);
 		return 0;
@@ -745,7 +738,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		}
 		error = gnbd_resend_requests(dev, SOCKET_I(inode));
 		if (error){
-			printk("quitting NBD_DO_IT\n");
+			printk("quitting GNBD_DO_IT\n");
 			up(&dev->do_it_lock);
 			fput(file);
 			return error;
@@ -770,7 +763,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		dev->file = NULL;
 		if (file)
 			fput(file);
-		printk("exitting GNBD_DO_IT ioctl\n");
+		printk("exiting GNBD_DO_IT ioctl\n");
 		return error;
 	case GNBD_PING:
 		/* FIXME -- should I allow pings if everything is compeletely
@@ -783,7 +776,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 			list_add(&ping_req.queuelist, &dev->queue_head);
 		}
 		spin_unlock(&dev->queue_lock);
-		gnbd_send_req(dev, &ping_req); /* ignore the errors */
+		gnbd_send_req(dev, &ping_req, 1); /* ignore the errors */
 		return 0;
 	case GNBD_PRINT_DEBUG:
 		printk(KERN_INFO "%s: next = %p, prev = %p, head = %p\n",
@@ -793,7 +786,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		return 0;
 	case GNBD_GET_TIME:
 		if (copy_to_user((void *)arg, &insmod_time, sizeof(uint64_t))){
-			printk(KERN_WARNING "couldn't compy time argument to user\n");
+			printk(KERN_WARNING "couldn't copy time argument to user\n");
 			return -EFAULT;
 		}
 		return 0;
@@ -884,10 +877,8 @@ static int __init gnbd_init(void)
 	struct timeval tv;
 	int i;
 
-	if (sizeof(struct gnbd_request) != 28) {
-		printk(KERN_CRIT "gnbd: sizeof gnbd_request needs to be 28 in order to work!\n" );
-		return -EIO;
-	}
+	BUILD_BUG_ON(sizeof(struct gnbd_request) != 28);
+
 	shutdown_req.cmd_type = REQ_TYPE_SPECIAL;
 	gnbd_cmd(&shutdown_req) = GNBD_CMD_DISC;
 	shutdown_req.sector = 0;
@@ -913,6 +904,13 @@ static int __init gnbd_init(void)
 			put_disk(disk);
 			goto out;
 		}
+		elevator_exit(disk->queue->elevator);
+		err = elevator_init(disk->queue, "anticipatory");
+		if (err) {
+			blk_cleanup_queue(disk->queue);
+			put_disk(disk);
+			goto out;
+		}
 	}
 	major_nr = register_blkdev(major_nr, "gnbd");
 	if (major_nr < 0) {
@@ -982,7 +980,7 @@ static int __init gnbd_init(void)
 		set_capacity(disk, 0);
 		add_disk(disk);
 		if(sysfs_create_link(&gnbd_dev[i].class_dev.kobj,
-					&gnbd_dev[i].disk->kobj, "block"))
+					&gnbd_dev[i].disk->dev.kobj, "block"))
 			goto out_remove_disk;
 		
 	}
@@ -991,6 +989,7 @@ static int __init gnbd_init(void)
         if (err)
                 goto out_unregister_class;
         
+	do_gettimeofday(&tv); 
 	insmod_time = (uint64_t) tv.tv_sec * 1000000 + tv.tv_usec;
 
 	return 0;
