@@ -1,7 +1,7 @@
 /******************************************************************************
 *******************************************************************************
 **
-**  Copyright (C) 2005 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2005-2007 Red Hat, Inc.  All rights reserved.
 **  
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -24,6 +24,7 @@
 #include <time.h>
 #include <syslog.h>
 #include <sched.h>
+#include <limits.h>
 #include <asm/types.h>
 #include <sys/socket.h>
 #include <sys/poll.h>
@@ -105,13 +106,12 @@ do { \
 } while (0)
 
 #define ASSERT(x) \
-{ \
+do { \
 	if (!(x)) { \
-		fprintf(stderr, "\nAssertion failed on line %d of file %s\n\n" \
-			"Assertion:  \"%s\"\n", \
-			__LINE__, __FILE__, #x); \
-        } \
-}
+		log_error("Assertion failed on line %d of file %s\n" \
+			  "Assertion:  \"%s\"\n", __LINE__, __FILE__, #x); \
+	} \
+} while (0)
 
 struct mountpoint {
 	struct list_head	list;
@@ -164,10 +164,12 @@ struct mountgroup {
 	int			low_nodeid;
 	int			master_nodeid;
 	int			save_plocks;
+	int			reject_mounts;
 
 	uint64_t		cp_handle;
 	time_t			last_checkpoint_time;
 	time_t			last_plock_time;
+	struct timeval		drop_resources_last;
 
 	int			needs_recovery;
 	int			our_jid;
@@ -241,12 +243,12 @@ enum {
 	MSG_MOUNT_STATUS,
 	MSG_RECOVERY_STATUS,
 	MSG_RECOVERY_DONE,
+	MSG_PLOCK_OWN,
+	MSG_PLOCK_DROP,
+	MSG_PLOCK_SYNC_LOCK,
+	MSG_PLOCK_SYNC_WAITER,
 };
 
-#define GDLM_VER_MAJOR 1
-#define GDLM_VER_MINOR 0
-#define GDLM_VER_PATCH 0
-
 struct gdlm_header {
 	uint16_t		version[3];
 	uint16_t		type;			/* MSG_ */
@@ -267,6 +269,9 @@ int do_read(int fd, void *buf, size_t count);
 int do_write(int fd, void *buf, size_t count);
 struct mountgroup *find_mg(char *name);
 struct mountgroup *find_mg_id(uint32_t id);
+struct mg_member *find_memb_nodeid(struct mountgroup *mg, int nodeid);
+int is_member(struct mountgroup *mg, int nodeid);
+int is_removed(struct mountgroup *mg, int nodeid);
 
 int setup_cman(void);
 int process_cman(void);
