@@ -19,7 +19,7 @@
 extern char *clustername;
 extern int our_nodeid;
 extern group_handle_t gh;
-extern int no_withdraw;
+extern int config_no_withdraw;
 extern int dmsetup_wait;
 
 struct list_head mounts;
@@ -1328,8 +1328,6 @@ void recover_members(struct mountgroup *mg, int num_nodes,
 				  memb->spectator,
 				  memb->wait_gfs_recover_done);
 
-			purge_plocks(mg, memb->nodeid, 0);
-
 			if (mg->master_nodeid == memb->nodeid &&
 			    memb->gone_type == GROUP_NODE_FAILED)
 				master_failed = 1;
@@ -1617,7 +1615,13 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 
 	mg = find_mg(name);
 	if (mg) {
-		rv = add_another_mountpoint(mg, dir, dev, ci);
+		if (mg->reject_mounts) {
+			/* fs is being unmounted */
+			rv = -ESTALE;
+			log_error("mount: reject mount due to unmount");
+		} else {
+			rv = add_another_mountpoint(mg, dir, dev, ci);
+		}
 		goto out;
 	}
 
@@ -1909,6 +1913,25 @@ int kernel_recovery_done(char *table)
 
 	switch (atoi(buf)) {
 	case LM_RD_GAVEUP:
+		/*
+		 * This is unfortunate; it's needed for bz 442451 where
+		 * gfs-kernel fails to acquire the journal lock on all nodes
+		 * because a withdrawing node has not yet called
+		 * dlm_release_lockspace() to free it's journal lock.  With
+		 * this, all nodes should repeatedly try to to recover the
+		 * journal of the withdrawn node until the withdrawing node
+		 * clears its dlm locks, and gfs on each of the remaining nodes
+		 * succeeds in doing the recovery.
+		 */
+
+		if (memb->withdrawing) {
+			log_group(mg, "recovery_done jid %d nodeid %d retry "
+				  "for withdraw", memb->jid, memb->nodeid);
+			memb->tell_gfs_to_recover = 1;
+			memb->wait_gfs_recover_done = 0;
+			usleep(500000);
+		}
+
 		memb->local_recovery_status = RS_GAVEUP;
 		ss = "gaveup";
 		break;
@@ -2030,6 +2053,7 @@ int do_unmount(int ci, char *dir, int mnterr)
 	}
 
  out:
+	mg->reject_mounts = 1;
 	group_leave(gh, mg->name);
 	return 0;
 }
@@ -2705,7 +2729,7 @@ int do_withdraw(char *table)
 	char *name = strstr(table, ":") + 1;
 	int rv;
 
-	if (no_withdraw) {
+	if (config_no_withdraw) {
 		log_error("withdraw feature not enabled");
 		return 0;
 	}
