@@ -50,8 +50,13 @@ static uint16_t _port = 0;		/** Our daemon ID, set with vf_init. */
  * TODO: We could make it thread safe, but this might be unnecessary work
  * Solution: Super-coarse-grained-bad-code-locking!
  */
+#ifdef WRAP_LOCKS
+static pthread_mutex_t key_list_mutex = PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP;
+static pthread_mutex_t vf_mutex = PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP;
+#else
 static pthread_mutex_t key_list_mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t vf_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif /* WRAP_LOCKS */
 static pthread_t vf_thread = (pthread_t)-1;
 static int vf_thread_ready = 0;
 static vf_vote_cb_t default_vote_cb = NULL;
@@ -116,9 +121,9 @@ static int vf_send_current(msgctx_t *, char *);
 
 
 struct vf_args {
-	uint16_t port;
-	int local_node_id;
 	msgctx_t *ctx;
+	int local_node_id;
+	uint16_t port;
 };
 
 
@@ -272,6 +277,9 @@ vf_resolve_views(key_node_t *key_node)
 	uint32_t datalen;
 	uint32_t trans;
 
+	if (!key_node)
+		return 0;
+
 	while ((trans = vf_try_commit(key_node)) != 0) {
 		commits++;
 	}
@@ -726,8 +734,13 @@ getuptime(struct timeval *tv)
 	if (!fp)
 		return -1;
 
+#if defined(__sparc__) || defined(__sparc64__)
+	rv = fscanf(fp,"%ld.%d %ld.%d\n", &tv->tv_sec, &tv->tv_usec,
+		    &junk.tv_sec, &junk.tv_usec);
+#else
 	rv = fscanf(fp,"%ld.%ld %ld.%ld\n", &tv->tv_sec, &tv->tv_usec,
 		    &junk.tv_sec, &junk.tv_usec);
+#endif
 	fclose(fp);
 	
 	if (rv != 4) {
@@ -788,6 +801,15 @@ vf_try_commit(key_node_t *key_node)
 		free(key_node->kn_data);
 	key_node->kn_datalen = vnp->vn_datalen;
 	key_node->kn_data = malloc(vnp->vn_datalen);
+
+	/*
+	 *   Need to check return of malloc always
+	 */
+	if (key_node->kn_data == NULL) {
+		fprintf (stderr, "malloc fail err=%d\n", errno);
+		return -1;
+	}
+
 	memcpy(key_node->kn_data, vnp->vn_data, vnp->vn_datalen);
 
 	free(vnp);
@@ -876,7 +898,7 @@ vf_server(void *arg)
 
 	msg_close(ctx);
 	msg_free_ctx(ctx);
-	return NULL;
+	pthread_exit(NULL);
 }
 
 
@@ -1013,6 +1035,13 @@ vf_key_init_nt(char *keyid, int timeout, vf_vote_cb_t vote_cb,
 	}
 
 	newnode = malloc(sizeof(*newnode));
+
+	if (newnode == NULL) {
+		fprintf(stderr, "malloc fail3 err=%d\n", errno);
+		pthread_mutex_unlock(&key_list_mutex);
+		return -1;
+	}
+
 	newnode->kn_data = NULL;
 	memset(newnode,0,sizeof(*newnode));
 	newnode->kn_keyid = strdup(keyid);
@@ -1136,7 +1165,6 @@ vf_write(cluster_member_list_t *membership, uint32_t flags, char *keyid,
 	snprintf(lock_name, sizeof(lock_name), "usrm::vf");
 	l = clu_lock(LKM_EXMODE, &lockp, 0, lock_name);
 	if (l < 0) {
-		clu_unlock(&lockp);
 		pthread_mutex_unlock(&vf_mutex);
 		return l;
 	}
@@ -1153,7 +1181,7 @@ vf_write(cluster_member_list_t *membership, uint32_t flags, char *keyid,
 	}
 
 #ifdef DEBUG
-	printf("aight, need responses from %d guys\n", remain);
+	printf("Allright, need responses from %d members\n", remain);
 #endif
 
 	pthread_mutex_lock(&key_list_mutex);
@@ -1195,6 +1223,8 @@ vf_write(cluster_member_list_t *membership, uint32_t flags, char *keyid,
 	 */
 	if (msg_open(MSG_CLUSTER, 0, _port, &everyone, 0) < 0) {
 		printf("msg_open: fail: %s\n", strerror(errno));
+		clu_unlock(&lockp);
+		pthread_mutex_unlock(&vf_mutex);
 		return -1;
 	}
 
@@ -1338,6 +1368,7 @@ vf_process_msg(msgctx_t *ctx, int nodeid, generic_msg_hdr *msgp, int nbytes)
 {
 	vf_msg_t *hdrp;
 	int ret;
+	key_node_t *kn;
 
 	if ((nbytes <= 0) || (nbytes < sizeof(generic_msg_hdr)) ||
 	    (msgp->gh_command != VF_MESSAGE))
@@ -1392,8 +1423,13 @@ vf_process_msg(msgctx_t *ctx, int nodeid, generic_msg_hdr *msgp, int nbytes)
 #endif
 		pthread_mutex_lock(&key_list_mutex);
 		vf_buffer_commit(msgp->gh_arg2);
-		ret = (vf_resolve_views(kn_find_trans(msgp->gh_arg2)) ?
-			VFR_COMMIT : VFR_OK);
+		kn = kn_find_trans(msgp->gh_arg2);
+		if (!kn) {
+			pthread_mutex_unlock(&key_list_mutex);
+			return VFR_OK;
+		}
+
+		ret = (vf_resolve_views(kn) ? VFR_COMMIT : VFR_OK);
 		pthread_mutex_unlock(&key_list_mutex);
 		return ret;
 
@@ -1432,9 +1468,7 @@ vf_read(cluster_member_list_t *membership, char *keyid, uint64_t *view,
 	snprintf(lock_name, sizeof(lock_name), "usrm::vf");
 	l = clu_lock(LKM_EXMODE, &lockp, 0, lock_name);
 	if (l < 0) {
-		clu_unlock(&lockp);
 		pthread_mutex_unlock(&vf_mutex);
-		printf("Couldn't lock %s\n", keyid);
 		return l;
 	}
 
@@ -1700,54 +1734,88 @@ vf_request_current(cluster_member_list_t *membership, char *keyid,
 		}
 		msg_close(&ctx);
 		msg = (vf_msg_t *)gh;
-		break;
-	}
 
-	if (x >= membership->cml_count)
-		return VFR_ERROR;
+		/* Uh oh */
+		if (!msg || (msg == &rmsg)) {
+			printf("VF: No valid message\n");
+			return VFR_ERROR;
+		}
+		swab_generic_msg_hdr(&(msg->vm_hdr));
+		if (msg->vm_hdr.gh_command == VF_NACK) {
+			free(msg);
+			continue;
+		}
+		if (msg->vm_hdr.gh_length < sizeof(vf_msg_t)) {
+			fprintf(stderr, "VF: Short reply from %d\n", x);
+			free(msg);
+			continue;
+		}
+		if (msg->vm_hdr.gh_length > n) {
+			fprintf(stderr,
+				"VF: Size mismatch during decode (%d > %d)\n",
+				msg->vm_hdr.gh_length, n);
+			free(msg);
+			continue;
+		}
 
-	/* Uh oh */
-	if (!msg || (msg == &rmsg)) {
-		printf("VF: No valid message\n");
-		return VFR_ERROR;
-	}
-		
-	swab_generic_msg_hdr(&(msg->vm_hdr));
-	if (msg->vm_hdr.gh_command == VF_NACK) {
-		free(msg);
-		return VFR_NODATA;
-	}
+		swab_vf_msg_info_t(&(msg->vm_msg));
 
-	if (msg->vm_hdr.gh_length < sizeof(vf_msg_t)) {
-		fprintf(stderr, "VF: Short reply from %d\n", x);
-		free(msg);
-		return VFR_ERROR;
-	}
+		if (msg->vm_msg.vf_datalen != (n - sizeof(*msg))) {
+			fprintf(stderr,"VF: Size mismatch during decode (\n");
+			free(msg);
+			continue;
+		}
 
-	if (msg->vm_hdr.gh_length > n) {
-		fprintf(stderr,"VF: Size mismatch during decode (%d > %d)\n",
-			msg->vm_hdr.gh_length, n);
-		free(msg);
-		return VFR_ERROR;
-	}
-
-	swab_vf_msg_info_t(&(msg->vm_msg));
-
-	if (msg->vm_msg.vf_datalen != (n - sizeof(*msg))) {
-		fprintf(stderr,"VF: Size mismatch during decode (\n");
-		free(msg);
-		return VFR_ERROR;
-	}
-
-	if (vf_set_current(keyid, msg->vm_msg.vf_view,
+		/* Ok... we've got data! */
+		if (vf_set_current(keyid, msg->vm_msg.vf_view,
 			   msg->vm_msg.vf_data,
 			   msg->vm_msg.vf_datalen) == VFR_ERROR) {
+			free(msg);
+			return VFR_ERROR;
+		}
+
 		free(msg);
-		return VFR_ERROR;
+
+		return VFR_OK;
 	}
 
-	free(msg);
-
-	return VFR_OK;
+	return VFR_NODATA;
 }
 
+
+void
+dump_vf_states(FILE *fp)
+{
+	key_node_t *cur;
+
+	fprintf(fp, "View-Formation States:\n");
+	fprintf(fp, "  Thread: %d\n", (unsigned)vf_thread);
+	fprintf(fp, "  Default callbacks:\n    Vote: %p\n    Commit: %p\n",
+		default_vote_cb, default_commit_cb);
+	fprintf(fp, "  Distributed key metadata:\n");
+
+	pthread_mutex_lock(&key_list_mutex);
+
+	for (cur = key_list; cur; cur = cur->kn_next) {
+		fprintf(fp, "    %s, View: %d, Size: %d, Address: %p\n",
+			cur->kn_keyid,
+			(int)cur->kn_viewno,
+			cur->kn_datalen,
+			cur->kn_data);
+		if (cur->kn_vote_cb != default_vote_cb) 
+			fprintf(fp, "      Vote callback: %p\n", cur->kn_vote_cb);
+		if (cur->kn_commit_cb != default_commit_cb) 
+			fprintf(fp, "      Commit callback: %p\n", cur->kn_commit_cb);
+
+		if (cur->kn_jvlist)
+			fprintf(fp, "        This key has unresolved "
+			        "new views pending\n");
+ 		if (cur->kn_clist)
+			fprintf(fp, "        This key has unresolved "
+			        "commits pending\n");
+
+	}
+
+	pthread_mutex_unlock(&key_list_mutex);
+	fprintf(fp, "\n");
+}
