@@ -444,22 +444,14 @@ findSlaves()
 		return $OCF_ERR_GENERIC
 	fi
 
-       ## Strip possible VLAN (802.1q) suffixes 
-       ##  - Roland Gadinger <roland.gadinger@beko.at> 
-       mastif=${mastif%%.*} 
+	## Strip possible VLAN (802.1q) suffixes 
+	##  - Roland Gadinger <roland.gadinger@beko.at> 
+	mastif=${mastif%%.*} 
 
 	while read line; do
 		set - $line
-		while [ $# -gt 0 ]; do
-			case $1 in
-			eth*:)
-				interfaces="${1/:/} $interfaces"
-				continue 2
-				;;
-			esac
-			shift
-		done
-	done < <( /sbin/ip link list | grep "master $mastif" )
+		interfaces="${2/:/} $interfaces"
+	done < <( /sbin/ip -o link list | grep "master $mastif" )
 
 	echo $interfaces
 }
@@ -692,6 +684,10 @@ ipv4()
 			if [ $? -ne 0 ]; then
 				continue
 			fi
+
+			if [ "${addr/\/*/}" = "${addr}" ]; then
+				addr="$addr/$maskbits"
+			fi
 			ocf_log info "Adding IPv4 address $addr to $dev"
 		fi
 		if [ "$1" = "del" ]; then
